@node XDR Technical Notes
@chapter XDR Technical Notes
@cindex XDR Technical Notes
@cindex XDR, Sun Technical Notes

This chapter contains technical notes on Sun's implementation of the
External Data Representation (XDR) standard, a set of library routines
that allow a C programmer to describe arbitrary data structures in a
machine-independent fashion.  For a formal specification of the XDR
standard, @xref{XDR Protocol Specification}.

@menu
* Overview::
* Justification::
* A Canonical Standard::
* The XDR Library::
* XDR Library Primitives::
* Advanced Topics::
@end menu

@node Overview
@section Overview
@cindex XDR, overview

This chapter contains technical notes on Sun's implementation of the
External Data Representation (@dfn{XDR}) standard, a set of library
routines that allow a C programmer to describe arbitrary data structures
in a machine-independent fashion.  For a formal specification of the XDR
standard, @xref{XDR Protocol Specification}.  XDR is the backbone of
Sun's Remote Procedure Call package, in the sense that data for remote
procedure calls is transmitted using the standard.  XDR library routines
should be used to transmit data that is accessed (read or written) by
more than one type of machine
@cindex XDR, system routines
@footnote{For a compete specification of the system External Data
Representation routines, see the @samp{xdr(3N)} manual page.}.

This chapter contains a short tutorial overview of the XDR library
routines, a guide to accessing currently available XDR streams, and
information on defining new streams and data types.  XDR was designed to
work across different languages, operating systems, and machine
architectures.  Most users (particularly RPC users) will only need the
information in the @ref{Number Filters}, @ref{Floating Point Filters},
and @ref{Enumeration Filters}.  Programmers wishing to implement RPC and
XDR on new machines will be interested in the rest of the chapter, as
well as the @ref{XDR Protocol Specification}, which will be their
primary reference.

@strong{Note:} @emph{ @code{rpcgen} can be used to write XDR routines
even in cases where no RPC calls are being made.}

On Sun systems, C programs that want to use XDR routines must include
the file @file{<rpc/rpc.h>}, which contains all the necessary interfaces
to the XDR system.  Since the C library @file{libc.a} contains all the
XDR routines, compile as normal:

@example
@cartouche
@b{example%} cc @i{program}.c
@end cartouche
@end example

@node Justification
@section Justification
@cindex XDR justification

Consider the following two programs, @code{writer}:

@example
#include <stdio.h>

main()                  /* @i{writer.c} */
@{
        long i;

        for (i = 0; i < 8; i++) @{
                if (fwrite((char *)&i, sizeof(i), 1, stdout) != 1) @{
                        fprintf(stderr, "failed!\n");
                        exit(1);
                @}
        @}
        exit(0);
@}
@end example

and @code{reader}:

@example
#include <stdio.h>

main()                  /* @i{reader.c} */
@{
        long i, j;

        for (j = 0; j < 8; j++) @{
                if (fread((char *)&i, sizeof (i), 1, stdin) != 1) @{
                        fprintf(stderr, "failed!\n");
                        exit(1);
                @}
                printf("%ld ", i);
        @}
        printf("\n");
        exit(0);
@}
@end example

The two programs appear to be portable, because

@enumerate a
@item 
they pass @code{lint} checking, and
@item
they exhibit the same behavior when executed on two different hardware
architectures, a Sun and a VAX.
@end enumerate

Piping the output of the @code{writer} program to the @code{reader}
program gives identical results on a Sun or a VAX.

@example
@cartouche
@b{sun%} writer | reader
0 1 2 3 4 5 6 7
@b{sun%}
@end cartouche

@cartouche
@b{vax%} writer | reader
0 1 2 3 4 5 6 7
@b{vax%}
@end cartouche
@end example

With the advent of local area networks and 4.2BSD came the concept of
``network pipes'' --- a process produces data on one machine, and a
second process consumes data on another machine.  A network pipe can be
constructed with @code{writer} and @code{reader}.  Here are the results
if the first produces data on a Sun, and the second consumes data on a
VAX.

@example
@cartouche
@b{sun%} writer | rsh vax reader
0 16777216 33554432 50331648 67108864 83886080 100663296 117440512
@b{sun%}
@end cartouche
@end example

Identical results can be obtained by executing @code{writer} on the VAX
and @code{reader} on the Sun.  These results occur because the byte
ordering of long integers differs between the VAX and the Sun, even
though word size is the same.  Note that 16777216 is 2^24 --- when four
bytes are reversed, the 1 winds up in the 24th bit.

Whenever data is shared by two or more machine types, there is a need
for portable data.  Programs can be made data-portable by replacing the
@code{read()} and @code{write()} calls with calls to an XDR library
routine @code{xdr_long()}, a filter that knows the standard
representation of a long integer in its external form.  Here are the
revised versions of @code{writer}:

@example
#include <stdio.h>
#include <rpc/rpc.h>    /* @i{xdr is a sub-library of rpc} */

main()          /* @i{writer.c} */
@{
        XDR xdrs;
        long i;

        xdrstdio_create(&xdrs, stdout, XDR_ENCODE);
        for (i = 0; i < 8; i++) @{
                if (!xdr_long(&xdrs, &i)) @{
                        fprintf(stderr, "failed!\n");
                        exit(1);
                @}
        @}
        exit(0);
@}
@end example

and @code{reader}:

@example
#include <stdio.h>
#include <rpc/rpc.h>    /* @i{xdr is a sub-library of rpc} */

main()          /* @i{reader.c} */
@{
        XDR xdrs;
        long i, j;

        xdrstdio_create(&xdrs, stdin, XDR_DECODE);
        for (j = 0; j < 8; j++) @{
                if (!xdr_long(&xdrs, &i)) @{
                        fprintf(stderr, "failed!\n");
                        exit(1);
                @}
                printf("%ld ", i);
        @}
        printf("\n");
        exit(0);
@}
@end example

The new programs were executed on a Sun, on a VAX, and from a Sun to a
VAX; the results are shown below.

@example
@cartouche
@b{sun%} writer | reader
0 1 2 3 4 5 6 7
@b{sun%}
@end cartouche

@cartouche
@b{vax%} writer | reader
0 1 2 3 4 5 6 7
@b{vax%}
@end cartouche

@cartouche
@b{sun%} writer | rsh vax reader
0 1 2 3 4 5 6 7
@b{sun%}
@end cartouche
@end example

@strong{Note:}
@cindex XDR, portable data
@emph{Integers are just the tip of the portable-data iceberg.
Arbitrary data structures present portability problems, particularly
with respect to alignment and pointers.  Alignment on word boundaries
may cause the size of a structure to vary from machine to machine.  And
pointers, which are very convenient to use, have no meaning outside the
machine where they are defined.}

@node A Canonical Standard
@section A Canonical Standard
@cindex XDR, canonical standard

XDR's approach to standardizing data representations is
@emph{canonical}.  That is, XDR defines a single byte order (Big
Endian), a single floating-point representation (IEEE), and so on.  Any
program running on any machine can use XDR to create portable data by
translating its local representation to the XDR standard
representations; similarly, any program running on any machine can read
portable data by translating the XDR standard representaions to its
local equivalents.  The single standard completely decouples programs
that create or send portable data from those that use or receive
portable data.  The advent of a new machine or a new language has no
effect upon the community of existing portable data creators and users.
A new machine joins this community by being ``taught'' how to convert
the standard representations and its local representations; the local
representations of other machines are irrelevant.  Conversely, to
existing programs running on other machines, the local representations
of the new machine are also irrelevant; such programs can immediately
read portable data produced by the new machine because such data
conforms to the canonical standards that they already understand.

There are strong precedents for XDR's canonical approach.  For example,
TCP/IP, UDP/IP, XNS, Ethernet, and, indeed, all protocols below layer
five of the ISO model, are canonical protocols.  The advantage of any
canonical approach is simplicity; in the case of XDR, a single set of
conversion routines is written once and is never touched again.  The
canonical approach has a disadvantage, but it is unimportant in
real-world data transfer applications.  Suppose two Little-Endian
machines are transferring integers according to the XDR standard.  The
sending machine converts the integers from Little-Endian byte order to
XDR (Big-Endian) byte order; the receiving machine performs the reverse
conversion.  Because both machines observe the same byte order, their
conversions are unnecessary.  The point, however, is not necessity, but
cost as compared to the alternative.

The time spent converting to and from a canonical representation is
insignificant, especially in networking applications.  Most of the time
required to prepare a data structure for transfer is not spent in
conversion but in traversing the elements of the data structure.  To
transmit a tree, for example, each leaf must be visited and each element
in a leaf record must be copied to a buffer and aligned there; storage
for the leaf may have to be deallocated as well.  Similarly, to receive
a tree, storage must be allocated for each leaf, data must be moved from
the buffer to the leaf and properly aligned, and pointers must be
constructed to link the leaves together.  Every machine pays the cost of
traversing and copying data structures whether or not conversion is
required.  In networking applications, communications overhead --- the
time required to move the data down through the sender's protocol
layers, across the network and up through the receiver's protocol layers
--- dwarfs conversion overhead.

@node The XDR Library
@section The XDR Library
@cindex XDR, library

The XDR library not only solves data portability problems, it also
allows you to write and read arbitrary C constructs in a consistent,
specified, well-documented manner.  Thus, it can make sense to use the
library even when the data is not shared among machines on a network.

The XDR library has filter routines for strings (null-terminated arrays
of bytes), structures, unions, and arrays, to name a few.  Using more
primitive routines, you can write your own specific XDR routines to
describe arbitrary data structures, including elements of arrays, arms
of unions, or objects pointed at from other structures.  The structures
themselves may contain arrays of arbitrary elements, or pointers to
other structures.

Let's examine the two programs more closely.  There is a family of XDR
stream creation routines in which each member treats the stream of bits
differently.  In our example, data is manipulated using standard I/O
routines, so we use @code{xdrstdio_create()}.
@iftex
@cindex @code{xdrstdio_create()}, example
@end iftex
@ifinfo
@cindex xdrstdio_create(), example
@end ifinfo

The parameters to XDR stream creation routines vary according to their
function.  In our example, @code{xdrstdio_create()} takes a pointer to
an XDR structure that it initializes, a pointer to a @code{FILE} that
the input or output is performed on, and the operation.  The operation
may be @code{XDR_ENCODE} for serializing in the @samp{writer} program,
or @code{XDR_DECODE} for deserializing in the @samp{reader} program.

Note: RPC users never need to create XDR streams; the RPC system itself
creates these streams, which are then passed to the users.

The @code{xdr_long()} primitive is characteristic of most XDR library
primitives and all client XDR routines.  First, the routine returns
@code{FALSE} (@code{0}) if it fails, and @code{TRUE} (@code{1}) if it
succeeds.  Second, for each data type, @code{xxx}, there is an
associated XDR routine of the form:

@example
xdr_xxx(xdrs, xp)
        XDR *xdrs;
        xxx *xp;
@{
@}
@end example

In our case, @code{xxx} is long, and the corresponding XDR routine is a
primitive, @code{xdr_long()}.  The client could also define an arbitrary
structure @code{xxx} in which case the client would also supply the
routine @code{xdr_xxx()}, describing each field by calling XDR routines
of the appropriate type.  In all cases the first parameter, @code{xdrs}
can be treated as an opaque handle, and passed to the primitive
routines.

XDR routines are direction independent; that is, the same routines are
called to serialize or deserialize data.  This feature is critical to
software engineering of portable data.  The idea is to call the same
routine for either operation --- this almost guarantees that serialized
data can also be deserialized.  One routine is used by both producer and
consumer of networked data.  This is implemented by always passing the
address of an object rather than the object itself --- only in the case
of deserialization is the object modified.  This feature is not shown in
our trivial example, but its value becomes obvious when nontrivial data
structures are passed among machines.  If needed, the user can obtain
the direction of the XDR operation.  See the @ref{XDR Operation
Directions} for details.

Let's look at a slightly more complicated example.  Assume that a
person's gross assets and liabilities are to be exchanged among
processes.  Also assume that these values are important enough to
warrant their own data type:

@example
struct gnumbers @{
        long g_assets;
        long g_liabilities;
@};
@end example

The corresponding XDR routine describing this structure would be:

@example
bool_t                  /* @i{TRUE is success, FALSE is failure} */
xdr_gnumbers(xdrs, gp)
        XDR *xdrs;
        struct gnumbers *gp;
@{
        if (xdr_long(xdrs, &gp->g_assets) &&
            xdr_long(xdrs, &gp->g_liabilities))
                return(TRUE);
        return(FALSE);
@}
@end example

Note that the parameter @code{xdrs} is never inspected or modified; it
is only passed on to the subcomponent routines.  It is imperative to
inspect the return value of each XDR routine call, and to give up
immediately and return @code{FALSE} if the subroutine fails.

This example also shows that the type @code{bool_t} is declared as an
integer whose only values are @code{TRUE} (1) and @code{FALSE} (0).
This document uses the following definitions:

@example
#define bool_t  int
#define TRUE    1
#define FALSE   0
@end example

Keeping these conventions in mind, @code{xdr_gnumbers()} can be
rewritten as follows:

@example
xdr_gnumbers(xdrs, gp)
        XDR *xdrs;
        struct gnumbers *gp;
@{
        return(xdr_long(xdrs, &gp->g_assets) &&
                xdr_long(xdrs, &gp->g_liabilities));
@}
@end example

This document uses both coding styles.

@node XDR Library Primitives
@section XDR Library Primitives
@cindex library primitives for XDR
@cindex XDR, library primitives

This section gives a synopsis of each XDR primitive.  It starts with
basic data types and moves on to constructed data types.  Finally, XDR
utilities are discussed.  The interface to these primitives and
utilities is defined in the include file @file{<rpc/xdr.h>},
automatically included by @file{<rpc/rpc.h>}.

@menu
* Number Filters::
* Floating Point Filters::
* Enumeration Filters::
* No Data::
* Constructed Data Type Filters::
* Non-filter Primitives::
* XDR Operation Directions::
* XDR Stream Access::
* XDR Stream Implementation::
@end menu

@node Number Filters
@subsection Number Filters
@cindex XDR library, number filters

The XDR library provides primitives to translate between numbers and
their corresponding external representations.  Primitives cover the set
of numbers in:

@example
[signed, unsigned] * [short, int, long]
@end example

Specifically, the eight primitives are:

@example
bool_t xdr_char(xdrs, cp)
        XDR *xdrs;
        char *cp;

bool_t xdr_u_char(xdrs, ucp)
        XDR *xdrs;
        unsigned char *ucp;

bool_t xdr_int(xdrs, ip)
        XDR *xdrs;
        int *ip;

bool_t xdr_u_int(xdrs, up)
        XDR *xdrs;
        unsigned *up;

bool_t xdr_long(xdrs, lip)
        XDR *xdrs;
        long *lip;

bool_t xdr_u_long(xdrs, lup)
        XDR *xdrs;
        u_long *lup;

bool_t xdr_short(xdrs, sip)
        XDR *xdrs;
        short *sip;

bool_t xdr_u_short(xdrs, sup)
        XDR *xdrs;
        u_short *sup;
@end example

The first parameter, @code{xdrs}, is an XDR stream handle.  The second
parameter is the address of the number that provides data to the stream
or receives data from it.  All routines return @code{TRUE} if they
complete successfully, and @code{FALSE} otherwise.

@node Floating Point Filters
@subsection Floating Point Filters
@cindex XDR library, floating point filters

The XDR library also provides primitive routines for C's floating point
types:

@example
bool_t xdr_float(xdrs, fp)
        XDR *xdrs;
        float *fp;

bool_t xdr_double(xdrs, dp)
        XDR *xdrs;
        double *dp;
@end example

The first parameter, @code{xdrs} is an XDR stream handle.  The second
parameter is the address of the floating point number that provides data
to the stream or receives data from it.  Both routines return
@code{TRUE} if they complete successfully, and @code{FALSE} otherwise.

Note: Since the numbers are represented in IEEE floating point, routines
may fail when decoding a valid IEEE representation into a
machine-specific representation, or vice-versa.

@node Enumeration Filters
@subsection Enumeration Filters
@cindex XDR library, enumeration filters

The XDR library provides a primitive for generic enumerations.  The
primitive assumes that a C @code{enum} has the same representation
inside the machine as a C integer.  The boolean type is an important
instance of the @code{enum}.  The external representation of a boolean
is always @code{TRUE} (1) or @code{FALSE} (0).

@example
#define bool_t  int
#define FALSE   0
#define TRUE    1

#define enum_t int

bool_t xdr_enum(xdrs, ep)
        XDR *xdrs;
        enum_t *ep;

bool_t xdr_bool(xdrs, bp)
        XDR *xdrs;
        bool_t *bp;
@end example

The second parameters @code{ep} and @code{bp} are addresses of the
associated type that provides data to, or receives data from, the stream
@code{xdrs}.

@node No Data
@subsection No Data
@cindex XDR library, no data

Occasionally, an XDR routine must be supplied to the RPC system, even
when no data is passed or required.  The library provides such a
routine:

@example
bool_t xdr_void();  /* @i{always returns TRUE} */
@end example

@node Constructed Data Type Filters
@subsection Constructed Data Type Filters
@cindex XDR library, constructed data type filters

Constructed or compound data type primitives require more parameters and
perform more complicated functions then the primitives discussed above.
This section includes primitives for strings, arrays, unions, and
pointers to structures.

Constructed data type primitives may use memory management.  In many
cases, memory is allocated when deserializing data with
@code{XDR_DECODE} Therefore, the XDR package must provide means to
deallocate memory.  This is done by an XDR operation, @code{XDR_FREE} To
review, the three XDR directional operations are @code{XDR_ENCODE},
@code{XDR_DECODE} and @code{XDR_FREE}.

@menu
* Strings::
* Byte Arrays::
* Arrays::
* XDR Examples::
* Opaque Data::
* Fixed Sized Arrays::
* Discriminated Unions::
* Pointers::
@end menu

@node Strings
@subsubsection Strings
@cindex XDR library, strings

In C, a string is defined as a sequence of bytes terminated by a null
byte, which is not considered when calculating string length.  However,
when a string is passed or manipulated, a pointer to it is employed.
Therefore, the XDR library defines a string to be a @code{char *} and
not a sequence of characters.  The external representation of a string
is drastically different from its internal representation.  Externally,
strings are represented as sequences of ASCII characters, while
internally, they are represented with character pointers.  Conversion
between the two representations is accomplished with the routine
@code{xdr_string()}:
@findex xdr_string()

@example
bool_t xdr_string(xdrs, sp, maxlength)
        XDR *xdrs;
        char **sp;
        u_int maxlength;
@end example

The first parameter @code{xdrs} is the XDR stream handle.  The second
parameter @code{sp} is a pointer to a string (type @code{char **}).  The
third parameter @code{maxlength} specifies the maximum number of bytes
allowed during encoding or decoding.  Its value is usually specified by
a protocol.  For example, a protocol specification may say that a file
name may be no longer than 255 characters.

The routine returns @code{FALSE} if the number of characters exceeds
@code{maxlength}, and @code{TRUE} if it doesn't.  @strong{Keep}
@code{maxlength} small.  If it is too big you can blow the heap, since
@code{xdr_string()} will call @code{malloc()} for space.

The behavior of @code{xdr_string()} is similar to the behavior of other
routines discussed in this section.  The direction @code{XDR_ENCODE} is
easiest to understand.  The parameter @code{sp} points to a string of a
certain length; if the string does not exceed @code{maxlength}, the
bytes are serialized.

The effect of deserializing a string is subtle.  First the length of the
incoming string is determined; it must not exceed @code{maxlength}.
Next @code{sp} is dereferenced; if the the value is @code{NULL}, then a
string of the appropriate length is allocated and @code{*sp} is set to
this string.  If the original value of @code{*sp} is non-null, then the
XDR package assumes that a target area has been allocated, which can
hold strings no longer than @code{maxlength}.  In either case, the
string is decoded into the target area.  The routine then appends a null
character to the string.

In the @code{XDR_FREE} operation, the string is obtained by
dereferencing @code{sp}.  If the string is not @code{NULL}, it is freed
and @code{*sp} is set to @code{NULL}.  In this operation,
@code{xdr_string()} ignores the @code{maxlength} parameter.

@node Byte Arrays
@subsubsection Byte Arrays
@cindex XDR library, byte arrays

Often variable-length arrays of bytes are preferable to strings.  Byte
arrays differ from strings in the following three ways:

@enumerate
@item
the length of the array (the byte count) is explicitly located in an
unsigned integer,
@item
the byte sequence is not terminated by a null character, and
@item
the external representation of the bytes is the same as their internal
representation.
@end enumerate

The primitive @code{xdr_bytes()}
@findex xdr_bytes() 
converts between the internal and external representations of byte
arrays:

@example
bool_t xdr_bytes(xdrs, bpp, lp, maxlength)
    XDR *xdrs;
    char **bpp;
    u_int *lp;
    u_int maxlength;
@end example

The usage of the first, second and fourth parameters are identical to
the first, second and third parameters of @code{xdr_string()},
respectively.  The length of the byte area is obtained by dereferencing
@code{lp} when serializing; @code{*lp} is set to the byte length when
deserializing.

@node Arrays
@subsubsection Arrays
@cindex XDR library, arrays

The XDR library package provides a primitive for handling arrays of
arbitrary elements.  The @code{xdr_bytes()} routine treats a subset of
generic arrays, in which the size of array elements is known to be 1,
and the external description of each element is built-in.  The generic
array primitive, @code{xdr_array()},
@findex xdr_array()
requires parameters identical to those of @code{xdr_bytes()} plus two
more: the size of array elements, and an XDR routine to handle each of
the elements.  This routine is called to encode or decode each element
of the array.

@example
bool_t
xdr_array(xdrs, ap, lp, maxlength, elementsiz, xdr_element)
    XDR *xdrs;
    char **ap;
    u_int *lp;
    u_int maxlength;
    u_int elementsiz;
    bool_t (*xdr_element)();
@end example

The parameter @code{ap} is the address of the pointer to the array.  If
@code{*ap} is @code{NULL} when the array is being deserialized, XDR
allocates an array of the appropriate size and sets @code{*ap} to that
array.  The element count of the array is obtained from @code{*lp} when
the array is serialized; @code{*lp} is set to the array length when the
array is deserialized.  The parameter @code{maxlength} is the maximum
number of elements that the array is allowed to have; @code{elementsiz}
is the byte size of each element of the array (the C function
@code{sizeof()} can be used to obtain this value).  The
@code{xdr_element()}
@findex xdr_element()
routine is called to serialize, deserialize, or free each element of the
array.

@node XDR Examples
@subsubsection XDR Examples
@cindex XDR library, constructed data type filters: examples

Before defining more constructed data types, it is appropriate to
present three examples.

@subsubheading Example A: Network User

A user on a networked machine can be identified by

@enumerate a
@item
the machine name, such as @samp{krypton}: see the @code{gethostname()}
man page;
@item
the user's UID: see the @code{geteuid()} man page; and
@item
the group numbers to which the user belongs: see the @code{getgroups()}
man page.
@end enumerate

A structure with this information and its associated XDR routine could
be coded like this:

@example
struct netuser @{
    char    *nu_machinename;
    int     nu_uid;
    u_int   nu_glen;
    int     *nu_gids;
@};
#define NLEN 255    /* @i{machine names < 256 chars} */
#define NGRPS 20    /* @i{user can't be in > 20 groups} */

bool_t
xdr_netuser(xdrs, nup)
    XDR *xdrs;
    struct netuser *nup;
@{
    return(xdr_string(xdrs, &nup->nu_machinename, NLEN) &&
        xdr_int(xdrs, &nup->nu_uid) &&
        xdr_array(xdrs, &nup->nu_gids, &nup->nu_glen, 
        NGRPS, sizeof (int), xdr_int));
@}
@end example

@subsubheading Example B: Party of Network Users

A party of network users could be implemented as an array of
@code{netuser} structure.  The declaration and its associated XDR
routines are as follows:

@example
struct party @{
    u_int p_len;
    struct netuser *p_nusers;
@};
#define PLEN 500    /* @i{max number of users in a party} */

bool_t
xdr_party(xdrs, pp)
    XDR *xdrs;
    struct party *pp;
@{
    return(xdr_array(xdrs, &pp->p_nusers, &pp->p_len, PLEN,
        sizeof (struct netuser), xdr_netuser));
@}
@end example

@subsubheading Example C: History of Commands

The well-known parameters to @code{main()}, @code{argc} and @code{argv}
can be combined into a structure.  An array of these structures can make
up a history of commands.  The declarations and XDR routines might look
like:

@example
struct cmd @{
    u_int c_argc;
    char **c_argv;
@};
#define ALEN 1000   /* @i{args cannot be > 1000 chars} */
#define NARGC 100   /* @i{commands cannot have > 100 args} */

struct history @{
    u_int h_len;
    struct cmd *h_cmds;
@};
#define NCMDS 75    /* @i{history is no more than 75 commands} */

bool_t
xdr_wrap_string(xdrs, sp)
    XDR *xdrs;
    char **sp;
@{
    return(xdr_string(xdrs, sp, ALEN));
@}

bool_t
xdr_cmd(xdrs, cp)
    XDR *xdrs;
    struct cmd *cp;
@{
    return(xdr_array(xdrs, &cp->c_argv, &cp->c_argc, NARGC,
        sizeof (char *), xdr_wrap_string));
@}

bool_t
xdr_history(xdrs, hp)
    XDR *xdrs;
    struct history *hp;
@{
    return(xdr_array(xdrs, &hp->h_cmds, &hp->h_len, NCMDS,
        sizeof (struct cmd), xdr_cmd));
@}
@end example

The most confusing part of this example is that the routine
@code{xdr_wrap_string()} is needed to package the @code{xdr_string()}
routine, because the implementation of @code{xdr_array()} only passes
two parameters to the array element description routine;
@code{xdr_wrap_string()} supplies the third parameter to
@code{xdr_string()}.

By now the recursive nature of the XDR library should be obvious.  Let's
continue with more constructed data types.

@node Opaque Data
@subsubsection Opaque Data
@cindex XDR library, opaque data

In some protocols, handles are passed from a server to client.  The
client passes the handle back to the server at some later time.  Handles
are never inspected by clients; they are obtained and submitted.  That
is to say, handles are opaque.  The @code{xdr_opaque()}
@findex xdr_opaque()
primitive is used for describing fixed sized, opaque bytes.

@example
bool_t xdr_opaque(xdrs, p, len)
    XDR *xdrs;
    char *p;
    u_int len;
@end example

The parameter @code{p} is the location of the bytes; @code{len} is the
number of bytes in the opaque object.  By definition, the actual data
contained in the opaque object are not machine portable.

@node Fixed Sized Arrays
@subsubsection Fixed Sized Arrays
@cindex XDR library, fixed sized arrays
@findex xdr_vector()

The XDR library provides a primitive, @code{xdr_vector()}, for
fixed-length arrays.

@example
#define NLEN 255    /* @i{machine names must be < 256 chars} */
#define NGRPS 20    /* @i{user belongs to exactly 20 groups} */

struct netuser @{
    char *nu_machinename;
    int nu_uid;
    int nu_gids[NGRPS];
@};

bool_t
xdr_netuser(xdrs, nup)
    XDR *xdrs;
    struct netuser *nup;
@{
    int i;

    if (!xdr_string(xdrs, &nup->nu_machinename, NLEN))
        return(FALSE);
    if (!xdr_int(xdrs, &nup->nu_uid))
        return(FALSE);
    if (!xdr_vector(xdrs, nup->nu_gids, NGRPS, sizeof(int), 
        xdr_int)) @{
            return(FALSE);
    @}
    return(TRUE);
@}
@end example

@node Discriminated Unions
@subsubsection Discriminated Unions
@cindex XDR library, discriminated unions
@findex xdr_union()

The XDR library supports discriminated unions.  A discriminated union is
a C union and an @code{enum_t} value that selects an ``arm'' of the
union.

@example
struct xdr_discrim @{
    enum_t value;
    bool_t (*proc)();
@};

bool_t xdr_union(xdrs, dscmp, unp, arms, defaultarm)
    XDR *xdrs;
    enum_t *dscmp;
    char *unp;
    struct xdr_discrim *arms;
    bool_t (*defaultarm)();  /* @i{may equal NULL} */
@end example

First the routine translates the discriminant of the union located at
@code{*dscmp}.  The discriminant is always an @code{enum_t}.  Next the
union located at @code{*unp} is translated.  The parameter @code{arms}
is a pointer to an array of @code{xdr_discrim} structures.  Each
structure contains an ordered pair of @emph{[value,proc]}.  If the
union's discriminant is equal to the associated @emph{value}, then the
@emph{proc} is called to translate the union.  The end of the
@code{xdr_discrim} structure array is denoted by a routine of value
@code{NULL} (@code{0}).  If the discriminant is not found in the
@code{arms} array, then the @code{defaultarm} procedure is called if it
is non-null; otherwise the routine returns @code{FALSE}.

@subsubheading Example D: Union

Suppose the type of a union may be integer, character pointer (a
string), or a @code{gnumbers} structure.  Also, assume the union and its
current type are declared in a structure.  The declaration is:

@example
enum utype @{ INTEGER=1, STRING=2, GNUMBERS=3 @};

struct u_tag @{
    enum utype utype;   /* @i{the union's discriminant} */
    union @{
        int ival;
        char *pval;
        struct gnumbers gn;
    @} uval;
@};
@end example

The following constructs and XDR procedure (de)serialize the
discriminated union:

@example
struct xdr_discrim u_tag_arms[4] = @{
    @{ INTEGER, xdr_int @},
    @{ GNUMBERS, xdr_gnumbers @}
    @{ STRING, xdr_wrap_string @},
    @{ __dontcare__, NULL @}
    /* @i{always terminate arms with a NULL xdr_proc} */
@}

bool_t
xdr_u_tag(xdrs, utp)
    XDR *xdrs;
    struct u_tag *utp;
@{
    return(xdr_union(xdrs, &utp->utype, &utp->uval,
        u_tag_arms, NULL));
@}
@end example

The routine @code{xdr_gnumbers()} was presented above in @ref{The XDR
Library}.  @code{xdr_wrap_string()} was presented in example C.  The
default @code{arm} parameter to @code{xdr_union()} (the last parameter)
is @code{NULL} in this example.  Therefore the value of the union's
discriminant may legally take on only values listed in the
@code{u_tag_arms} array.  This example also demonstrates that the
elements of the arm's array do not need to be sorted.

It is worth pointing out that the values of the discriminant may be
sparse, though in this example they are not.  It is always good practice
to assign explicitly integer values to each element of the
discriminant's type.  This practice both documents the external
representation of the discriminant and guarantees that different C
compilers emit identical discriminant values.

Exercise: Implement @code{xdr_union()} using the other primitives in
this section.

@node Pointers
@subsubsection Pointers
@cindex XDR library, pointers

In C it is often convenient to put pointers to another structure within
a structure.  The @code{xdr_reference()}
@findex xdr_reference()
primitive makes it easy to serialize, deserialize, and free these
referenced structures.

@example
bool_t xdr_reference(xdrs, pp, size, proc)
    XDR *xdrs;
    char **pp;
    u_int ssize;
    bool_t (*proc)();
@end example

Parameter @code{pp} is the address of the pointer to the structure;
parameter @code{ssize} is the size in bytes of the structure (use the C
function @code{sizeof()} to obtain this value); and @code{proc} is the
XDR routine that describes the structure.  When decoding data, storage
is allocated if @code{*pp} is @code{NULL}.

There is no need for a primitive @code{xdr_struct()} to describe
structures within structures, because pointers are always sufficient.

Exercise: Implement @code{xdr_reference()} using @code{xdr_array()}.
Warning: @code{xdr_reference()} and @code{xdr_array()} are NOT
interchangeable external representations of data.

@subsubheading Example E: Pointers

Suppose there is a structure containing a person's name and a pointer to
a @code{gnumbers} structure containing the person's gross assets and
liabilities.  The construct is:

@example
struct pgn @{
    char *name;
    struct gnumbers *gnp;
@};
@end example

The corresponding XDR routine for this structure is:

@example
bool_t
xdr_pgn(xdrs, pp)
    XDR *xdrs;
    struct pgn *pp;
@{
    if (xdr_string(xdrs, &pp->name, NLEN) &&
      xdr_reference(xdrs, &pp->gnp,
      sizeof(struct gnumbers), xdr_gnumbers))
        return(TRUE);
    return(FALSE);
@}
@end example

@cindex Pointer semantics and XDR

@subsubheading Pointer Semantics and XDR

In many applications, C programmers attach double meaning to the values
of a pointer.  Typically the value @code{NULL} (or zero) means data is
not needed, yet some application-specific interpretation applies.  In
essence, the C programmer is encoding a discriminated union efficiently
by overloading the interpretation of the value of a pointer.  For
instance, in example E a @code{NULL} pointer value for @code{gnp} could
indicate that the person's assets and liabilities are unknown.  That is,
the pointer value encodes two things: whether or not the data is known;
and if it is known, where it is located in memory.  Linked lists are an
extreme example of the use of application-specific pointer
interpretation.

The primitive @code{xdr_reference()}
@iftex
@cindex @code{xdr_reference()} with NULL pointers
@end iftex
@ifinfo
@cindex xdr_reference() with NULL pointers
@end ifinfo
cannot and does not attach any special meaning to a null-value pointer
during serialization.  That is, passing an address of a pointer whose
value is @code{NULL} to @code{xdr_reference()} when serialing data will
most likely cause a memory fault and, on the UNIX system, a core dump.

@code{xdr_pointer()} correctly handles @code{NULL} pointers.  For more
information about its use, see the @ref{Linked Lists}.

@subsubheading Exercise:

After reading the @ref{Linked Lists}, return here and extend example E
so that it can correctly deal with @code{NULL} pointer values.

@subsubheading Exercise:

Using the @code{xdr_union()}, @code{xdr_reference()} and
@code{xdr_void()} primitives, implement a generic pointer handling
primitive that implicitly deals with @code{NULL} pointers.  That is,
implement @code{xdr_pointer()}.

@node Non-filter Primitives
@subsection Non-filter Primitives
@cindex XDR, non-filter primitives

XDR streams can be manipulated with the primitives discussed in this
section.

@example
u_int xdr_getpos(xdrs)
    XDR *xdrs;

bool_t xdr_setpos(xdrs, pos)
    XDR *xdrs;
    u_int pos;

xdr_destroy(xdrs)
    XDR *xdrs;
@end example

The routine @code{xdr_getpos()}
@findex xdr_getpos()
returns an unsigned integer that describes the current position in the
data stream.  Warning: In some XDR streams, the returned value of
@code{xdr_getpos()} is meaningless; the routine returns a -1 in this
case (though -1 should be a legitimate value).

The routine @code{xdr_setpos()}
@findex xdr_setpos()
sets a stream position to @code{pos}.  Warning: In some XDR streams,
setting a position is impossible; in such cases, @code{xdr_setpos()}
will return @code{FALSE}.  This routine will also fail if the requested
position is out-of-bounds.  The definition of bounds varies from stream
to stream.

The @code{xdr_destroy()}
@findex xdr_destroy()
primitive destroys the XDR stream.  Usage of the stream after calling
this routine is undefined.

@node XDR Operation Directions
@subsection XDR Operation Directions
@cindex XDR, operation directions
@cindex direction of XDR operations

At times you may wish to optimize XDR routines by taking advantage of
the direction of the operation --- @code{XDR_ENCODE}, @code{XDR_DECODE}
or @code{XDR_FREE}.  The value @code{xdrs->x_op} always contains the
direction of the XDR operation.  Programmers are not encouraged to take
advantage of this information.  Therefore, no example is presented here.
However, an example in the @ref{Linked Lists}, demonstrates the
usefulness of the @code{xdrs->x_op} field.

@node XDR Stream Access
@subsection XDR Stream Access
@cindex XDR, stream access

An XDR stream is obtained by calling the appropriate creation routine.
These creation routines take arguments that are tailored to the specific
properties of the stream.

Streams currently exist for (de)serialization of data to or from
standard I/O @code{FILE} streams, TCP/IP connections and UNIX files, and
memory.

@menu
* Standard I/O Streams::
* Memory Streams::
* Record (TCP/IP) Streams::
@end menu

@node Standard I/O Streams
@subsubsection Standard I/O Streams
@cindex XDR, standard I/O streams

XDR streams can be interfaced to standard I/O using the
@code{xdrstdio_create()}
@findex xdrstdio_create()
routine as follows:

@example
#include <stdio.h>
#include <rpc/rpc.h>    /* @i{xdr streams part of rpc} */

void
xdrstdio_create(xdrs, fp, x_op)
    XDR *xdrs;
    FILE *fp;
    enum xdr_op x_op;
@end example

The routine @code{xdrstdio_create()} initializes an XDR stream pointed
to by @code{xdrs}.  The XDR stream interfaces to the standard I/O
library.  Parameter @code{fp} is an open file, and @code{x_op} is an XDR
direction.

@node Memory Streams
@subsubsection Memory Streams
@cindex XDR, memory streams

Memory streams allow the streaming of data into or out of a specified
area of memory:

@example
#include <rpc/rpc.h>

void
xdrmem_create(xdrs, addr, len, x_op)
    XDR *xdrs;
    char *addr;
    u_int len;
    enum xdr_op x_op;
@end example

The routine @code{xdrmem_create()}
@findex xdrmem_create()
initializes an XDR stream in local memory.  The memory is pointed to by
parameter @code{addr}; parameter @code{len} is the length in bytes of
the memory.  The parameters @code{xdrs} and @code{x_op} are identical to
the corresponding parameters of @code{xdrstdio_create()}.  Currently,
the UDP/IP implementation of RPC uses @code{xdrmem_create()}.  Complete
call or result messages are built in memory before calling the
@code{sendto()} system routine.

@node Record (TCP/IP) Streams
@subsubsection Record (TCP/IP) Streams
@cindex XDR, record (TCP/IP) streams

A record stream is an XDR stream built on top of a record marking
standard that is built on top of the UNIX file or 4.2 BSD connection
interface.

@example
#include <rpc/rpc.h>    /* @i{xdr streams part of rpc} */

xdrrec_create(xdrs,
  sendsize, recvsize, iohandle, readproc, writeproc)
    XDR *xdrs;
    u_int sendsize, recvsize;
    char *iohandle;
    int (*readproc)(), (*writeproc)();
@end example

The routine @code{xdrrec_create()} provides an XDR stream interface that
allows for a bidirectional, arbitrarily long sequence of records.  The
contents of the records are meant to be data in XDR form.  The stream's
primary use is for interfacing RPC to TCP connections.  However, it can
be used to stream data into or out of normal UNIX files.

The parameter @code{xdrs} is similar to the corresponding parameter
described above.  The stream does its own data buffering similar to that
of standard I/O.  The parameters @code{sendsize} and @code{recvsize}
determine the size in bytes of the output and input buffers,
respectively; if their values are zero (@code{0}), then predetermined
defaults are used.  When a buffer needs to be filled or flushed, the
routine @code{readproc()} or @code{writeproc()} is called, respectively.
The usage and behavior of these routines are similar to the UNIX system
calls @code{read()} and @code{write()}.  However, the first parameter to
each of these routines is the opaque parameter @code{iohandle}.  The
other two parameters @code{buf} and @code{nbytes}) and the results (byte
count) are identical to the system routines.  If @code{xxx} is
@code{readproc()} or @code{writeproc()}, then it has the following form:

@example
/*
 * @i{returns the actual number of bytes transferred.}
 * @i{-1 is an error}
 */
int
xxx(iohandle, buf, len)
    char *iohandle;
    char *buf;
    int nbytes;
@end example

The XDR stream provides means for delimiting records in the byte stream.
The implementation details of delimiting records in a stream are
discussed in the @ref{Advanced Topics}.  The primitives that are
specific to record streams are as follows:

@example
bool_t
xdrrec_endofrecord(xdrs, flushnow)
    XDR *xdrs;
    bool_t flushnow;

bool_t
xdrrec_skiprecord(xdrs)
    XDR *xdrs;

bool_t
xdrrec_eof(xdrs)
    XDR *xdrs;
@end example

The routine @code{xdrrec_endofrecord()}
@findex xdrrec_endofrecord()
causes the current outgoing data to be marked as a record.  If the
parameter @code{flushnow} is @code{TRUE}, then the stream's
@code{writeproc} will be called; otherwise, @code{writeproc} will be
called when the output buffer has been filled.

The routine @code{xdrrec_skiprecord()}
@findex xdrrec_skiprecord()
causes an input stream's position to be moved past the current record
boundary and onto the beginning of the next record in the stream.

If there is no more data in the stream's input buffer, then the routine
@code{xdrrec_eof()}
@findex xdrrec_eof()
returns @code{TRUE}.  That is not to say that there is no more data in
the underlying file descriptor.

@node XDR Stream Implementation
@subsection XDR Stream Implementation
@cindex XDR, stream implementation
@cindex stream implementation in XDR

This section provides the abstract data types needed to implement new
instances of XDR streams.

@menu
* The XDR Object::
@end menu

@node The XDR Object
@subsubsection The XDR Object
@cindex XDR object

The following structure defines the interface to an XDR stream:

@example
enum xdr_op @{ XDR_ENCODE=0, XDR_DECODE=1, XDR_FREE=2 @};

typedef struct @{
    enum xdr_op x_op;            /* @i{operation; fast added param} */
    struct xdr_ops @{
        bool_t  (*x_getlong)();  /* @i{get long from stream} */
        bool_t  (*x_putlong)();  /* @i{put long to stream} */
        bool_t  (*x_getbytes)(); /* @i{get bytes from stream} */
        bool_t  (*x_putbytes)(); /* @i{put bytes to stream} */
        u_int   (*x_getpostn)(); /* @i{return stream offset} */
        bool_t  (*x_setpostn)(); /* @i{reposition offset} */
        caddr_t (*x_inline)();   /* @i{ptr to buffered data} */
        VOID    (*x_destroy)();  /* @i{free private area} */
    @} *x_ops;
    caddr_t     x_public;        /* @i{users' data} */
    caddr_t     x_private;       /* @i{pointer to private data} */
    caddr_t     x_base;          /* @i{private for position info} */
    int         x_handy;         /* @i{extra private word} */
@} XDR;
@end example

The @code{x_op} field is the current operation being performed on the
stream.  This field is important to the XDR primitives, but should not
affect a stream's implementation.  That is, a stream's implementation
should not depend on this value.  The fields @code{x_private},
@code{x_base}, and @code{x_handy} are private to the particular stream's
implementation.  The field @code{x_public} is for the XDR client and
should never be used by the XDR stream implementations or the XDR
primitives.  @code{x_getpostn()}, @code{x_setpostn()} and
@code{x_destroy()} are macros for accessing operations.  The operation
@code{x_inline()} takes two parameters: an @code{XDR *}, and an unsigned
integer, which is a byte count.  The routine returns a pointer to a
piece of the stream's internal buffer.  The caller can then use the
buffer segment for any purpose.  From the stream's point of view, the
bytes in the buffer segment have been consumed or put.  The routine may
return @code{NULL} if it cannot return a buffer segment of the requested
size.  (The @code{x_inline()} routine is for cycle squeezers.  Use of
the resulting buffer is not data-portable.  Users are encouraged not to
use this feature.)

The operations @code{x_getbytes()} and @code{x_putbytes()} blindly get
and put sequences of bytes from or to the underlying stream; they return
@code{TRUE} if they are successful, and @code{FALSE} otherwise.  The
routines have identical parameters (replace @code{xxx}):

@example
bool_t
xxxbytes(xdrs, buf, bytecount)
        XDR *xdrs;
        char *buf;
        u_int bytecount;
@end example

The operations @code{x_getlong()} and @code{x_putlong()} receive and put
long numbers from and to the data stream.  It is the responsibility of
these routines to translate the numbers between the machine
representation and the (standard) external representation.  The UNIX
primitives @code{htonl()} and @code{ntohl()} can be helpful in
accomplishing this.  The higher-level XDR implementation assumes that
signed and unsigned long integers contain the same number of bits, and
that nonnegative integers have the same bit representations as unsigned
integers.  The routines return @code{TRUE} if they succeed, and
@code{FALSE} otherwise.  They have identical parameters:

@example
bool_t
xxxlong(xdrs, lp)
        XDR *xdrs;
        long *lp;
@end example

Implementors of new XDR streams must make an XDR structure (with new
operation routines) available to clients, using some kind of create
routine.

@node Advanced Topics
@section Advanced Topics
@cindex XDR, advanced topics

This section describes techniques for passing data structures that are
not covered in the preceding sections.  Such structures include linked
lists (of arbitrary lengths).  Unlike the simpler examples covered in
the earlier sections, the following examples are written using both the
XDR C library routines and the XDR data description language.  The
@ref{XDR Protocol Specification}
describes this language in complete detail.

@menu
* Linked Lists::
@end menu

@node Linked Lists
@subsection Linked Lists
@cindex XDR, linked lists

The last example in the @ref{Pointers} topic earlier in this chapter
presented a C data structure and its associated XDR routines for a
individual's gross assets and liabilities.  The example is duplicated
below:

@example
struct gnumbers @{
        long g_assets;
        long g_liabilities;
@};

bool_t
xdr_gnumbers(xdrs, gp)
        XDR *xdrs;
        struct gnumbers *gp;
@{
        if (xdr_long(xdrs, &(gp->g_assets)))
                return(xdr_long(xdrs, &(gp->g_liabilities)));
        return(FALSE);
@}
@end example

Now assume that we wish to implement a linked list of such information.
A data structure could be constructed as follows:

@example
struct gnumbers_node @{
        struct gnumbers gn_numbers;
        struct gnumbers_node *gn_next;
@};

typedef struct gnumbers_node *gnumbers_list;
@end example

The head of the linked list can be thought of as the data object; that
is, the head is not merely a convenient shorthand for a structure.
Similarly the @code{gn_next} field is used to indicate whether or not
the object has terminated.  Unfortunately, if the object continues, the
@code{gn_next} field is also the address of where it continues. The link
addresses carry no useful information when the object is serialized.

The XDR data description of this linked list is described by the
recursive declaration of @code{gnumbers_list}:

@example
struct gnumbers @{
        int g_assets;
        int g_liabilities;
@};

struct gnumbers_node @{
        gnumbers gn_numbers;
        gnumbers_node *gn_next;
@};
@end example

In this description, the boolean indicates whether there is more data
following it. If the boolean is @code{FALSE}, then it is the last data
field of the structure. If it is @code{TRUE}, then it is followed by a
gnumbers structure and (recursively) by a @code{gnumbers_list}.  Note
that the C declaration has no boolean explicitly declared in it (though
the @code{gn_next} field implicitly carries the information), while the
XDR data description has no pointer explicitly declared in it.

Hints for writing the XDR routines for a @code{gnumbers_list} follow
easily from the XDR description above. Note how the primitive
@code{xdr_pointer()} is used to implement the XDR union above.

@example
bool_t
xdr_gnumbers_node(xdrs, gn)
        XDR *xdrs;
        gnumbers_node *gn;
@{
        return(xdr_gnumbers(xdrs, &gn->gn_numbers) &&
                xdr_gnumbers_list(xdrs, &gp->gn_next));
@}

bool_t
xdr_gnumbers_list(xdrs, gnp)
        XDR *xdrs;
        gnumbers_list *gnp;
@{
        return(xdr_pointer(xdrs, gnp, 
                sizeof(struct gnumbers_node), 
                xdr_gnumbers_node));
@}
@end example

The unfortunate side effect of XDR'ing a list with these routines is
that the C stack grows linearly with respect to the number of node in
the list.  This is due to the recursion. The following routine collapses
the above two mutually recursive into a single, non-recursive one.

@example
bool_t
xdr_gnumbers_list(xdrs, gnp)
        XDR *xdrs;
        gnumbers_list *gnp;
@{
        bool_t more_data;
        gnumbers_list *nextp;

        for (;;) @{
                more_data = (*gnp != NULL);
                if (!xdr_bool(xdrs, &more_data)) @{
                        return(FALSE);
                @}
                if (! more_data) @{
                        break;
                @}
                if (xdrs->x_op == XDR_FREE) @{
                        nextp = &(*gnp)->gn_next;
                @}
                if (!xdr_reference(xdrs, gnp, 
                        sizeof(struct gnumbers_node), xdr_gnumbers)) @{

                return(FALSE);
                @}
                gnp = (xdrs->x_op == XDR_FREE) ? 
                        nextp : &(*gnp)->gn_next;
        @}
        *gnp = NULL;
        return(TRUE);
@}
@end example

The first task is to find out whether there is more data or not, so that
this boolean information can be serialized. Notice that this statement
is unnecessary in the @code{XDR_DECODE} case, since the value of
@code{more_data} is not known until we deserialize it in the next
statement.

The next statement XDR's the more_data field of the XDR union.  Then if
there is truly no more data, we set this last pointer to @code{NULL} to
indicate the end of the list, and return @code{TRUE} because we are
done. Note that setting the pointer to @code{NULL} is only important in
the @code{XDR_DECODE} case, since it is already @code{NULL} in the
@code{XDR_ENCODE} and @code{XDR_FREE} cases.

Next, if the direction is @code{XDR_FREE}, the value of @code{nextp} is
set to indicate the location of the next pointer in the list.  We do
this now because we need to dereference gnp to find the location of the
next item in the list, and after the next statement the storage pointed
to by @code{gnp} will be freed up and no be longer valid.  We can't do
this for all directions though, because in the @code{XDR_DECODE}
direction the value of @code{gnp} won't be set until the next statement.

Next, we XDR the data in the node using the primitive
@code{xdr_reference()}.  @code{xdr_reference()} is like
@code{xdr_pointer()} which we used before, but it does not send over the
boolean indicating whether there is more data.  We use it instead of
@code{xdr_pointer()} because we have already XDR'd this information
ourselves. Notice that the xdr routine passed is not the same type as an
element in the list. The routine passed is @code{xdr_gnumbers()}, for
XDR'ing gnumbers, but each element in the list is actually of type
@code{gnumbers_node}.  We don't pass @code{xdr_gnumbers_node()} because
it is recursive, and instead use @code{xdr_gnumbers()} which XDR's all
of the non-recursive part.  Note that this trick will work only if the
@code{gn_numbers} field is the first item in each element, so that their
addresses are identical when passed to @code{xdr_reference()}.

Finally, we update @code{gnp} to point to the next item in the list. If
the direction is @code{XDR_FREE}, we set it to the previously saved
value, otherwise we can dereference @code{gnp} to get the proper value.
Though harder to understand than the recursive version, this
non-recursive routine is far less likely to blow the C stack.  It will
also run more efficiently since a lot of procedure call overhead has
been removed. Most lists are small though (in the hundreds of items or
less) and the recursive version should be sufficient for them.

